/*
17-5 Lagged Fibonacci random number generator 
with thread independent generators
by Sigismondo Boschi
Agoust 13th, 2009
to be checked! 
Please send feedback if you find any error or realize some improvement
... and either if you check it against some reference and find it correct!
net1701@gmail.com
*/

#ifndef _HELLOK_H_
#define _HELLOK_H_

#define MAX_THREADS_PER_BLOCK 32
#define LFG_REG_SIZE 18

__shared__ unsigned int lfgGReg[17*MAX_THREADS_PER_BLOCK];
__shared__ volatile int lfgL[MAX_THREADS_PER_BLOCK];
__shared__ volatile int lfgK[MAX_THREADS_PER_BLOCK];

/* 
Each thread initialize its state in global memory. This does not need to be opitimized
so each thread occupy an adjacent block of data.
The initialization SHOULD guarantee that given a seed, all the nthreads gets seeds in the range

   [iseed, iseed+nthrads-1]
   
using a particular skeme of initialization, based on the canonical form of the lagged fibonacci generator
this guarantee that every thread will have its own sequence.
*/
__global__ void LFG175gmInit (int iseed, unsigned int *gfibreg) 
{
	int ig = threadIdx.x +  blockIdx.x*blockDim.x;
	unsigned int *fibreg = gfibreg + ig* LFG_REG_SIZE;
	const int a=16807;
	const int r=2836; 
	const int q=127773;
	iseed += ig;
	for(int i=0; i<16; i++) {
		int hi=iseed/q;
		int lo=iseed % q;
		iseed = a*lo - r*hi;
		if (iseed < 0) iseed += 0x7FFFFFFF;
		fibreg[i] = iseed << 1; 
    }
	fibreg[10] |= 1;
	fibreg[16] = 0;
	// one more to store the current index. Initial condition:
	// l = 16, k = 4
	fibreg[17] = 16;
}

/* 
Each thread copies its data from global to shared memory. 
Avoid bank conflicts letting and interleaved storage in shared memory 
*/
__device__ void LFG175gmLoadToShm (unsigned int *fibreg) 
{
	int ig = threadIdx.x +  blockIdx.x*blockDim.x;
	unsigned int *m = fibreg +  LFG_REG_SIZE*ig;	
	for(int i=0;i<17;i++) {
		lfgGReg[threadIdx.x+i*blockDim.x] = m[i];
	}
	lfgL[threadIdx.x] = m[17];
	lfgK[threadIdx.x] = (17+m[17]-12)%17;
}

/* 
Each thread copies its data from shared to global memory.
Useful to save the generator state to disk.
Note that you need to save the L index value as well, m[17]
so the values are 18, not 17. 
*/
__device__ void LFG175gmSaveFromShm (unsigned int *fibreg) 
{
	int ig = threadIdx.x +  blockIdx.x*blockDim.x;
	unsigned int *m = fibreg +  LFG_REG_SIZE*ig;	
	for(int i=0;i<17;i++) {
		m[i] = lfgGReg[threadIdx.x+i*blockDim.x];
	}
	m[17] = lfgL[threadIdx.x];
}

/*
Lagged Fibonacci core routine.
The generated number is divided by 2 because the LSB is the same for all the sequences, 
and better to discard it at all:
all the sequences in the canonical form have it set the same, for all the register.	
So we are generating integer signed values in the range 

   [ 0 - 0x7FFFFFFF ]
   
*/
__device__ int LFG175shm() {
	volatile unsigned int l = lfgL[threadIdx.x];
	volatile unsigned int k = lfgK[threadIdx.x];
	unsigned int r = lfgGReg[threadIdx.x+l*blockDim.x] + lfgGReg[threadIdx.x+k*blockDim.x];  
	lfgGReg[threadIdx.x+l*blockDim.x] = r;
	l = l==0 ? 16 : l-1;
	k = k==0 ? 16 : k-1;
	lfgL[threadIdx.x]=l;
	lfgK[threadIdx.x]=k;
	return r>>1;
}

/* 
usage example. LFG175gmInit must be called before, on host, or eventually you can start from
a previously saved state. 
 */
__global__ void LFG175test(unsigned int *gfibreg, int *out) 
{
	int ig = threadIdx.x +  blockIdx.x*blockDim.x;
	int *m = out + 100*ig;
	LFG175gmLoadToShm (gfibreg);
	for (int i=0; i < 100; i++) 
	{
		m[i] = LFG175shm();
	}
	LFG175gmSaveFromShm(gfibreg);
}

#endif // #ifndef _HELLOK_H_
